package jais;

import jais.communication.*;
import jais.utils.*;
import jais.world.*;
import jais.objects.InitInfo;
import jais.visualization.*;
import jais.player.*;

import org.apache.log4j.*;
import java.io.*;
import java.util.*;

/**
 * <p>Title: JAIS - JAVA Implemented Soccer</p>
 * <p>Description: This is the main class of this Robocup-Agent. Here everything goes together</p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: JAIS</p>
 * <p></p>
 * <p>for help type</p>
 * <p>java jais.Jais -ht</p>
 * <p></p>
 * @author Stephan Diederich
 */
public class Jais extends Thread
{
  static Logger logger=Logger.getLogger("jais");

  JaisTeam m_jaisTeam; //reference on our creator if present
  WorldModel m_worldModel;
  Visualization m_visualization;
  KI m_ki;
  Communication m_communication;
  String m_teamName;
  int m_playerNumber;
  boolean m_running = false;
  boolean m_visualizationEnabled=false;
  Thread m_thisThread=null; //reference to ourself
  boolean m_threadSuspended=false; //is our thread currently suspended?
  LinkedList m_eventList;
  boolean m_commandLineReading=true;
  boolean m_singleMode=false;
  BufferedReader m_inputReader; //needed for reading from commandline if in singlemode
  /**
   * @param f_teamName name of our team
   * @param f_serverName name of the server we want to connect to
   * @param f_serverPort port of the server we want to connect to
   * @param f_isGoalie true if we want a goalie to connect
   * @param f_playerNumber number of player we want to connect
   * @param f_singleMode if we are in singleMode, we have to start commandlineparsing
   *
   * @throws xJais if something with the initialization went wrong
   * @roseuid 3CE6898C0366
   */
  public Jais(JaisTeam f_jaisTeam, String f_teamName, String f_serverName, int f_serverPort, boolean f_isGoalie,int f_playerNumber, boolean f_singleMode) throws xJais
  {
    m_jaisTeam=f_jaisTeam;
    this.setName("Jais"+f_playerNumber);
    m_teamName=f_teamName; //store our Teamname
    m_playerNumber=f_playerNumber;
    try
    {
      m_communication=new Communication(m_teamName,f_serverName, f_serverPort,  f_isGoalie,this,f_playerNumber);
    }
    catch (xCommunication xCom)
    {
      logger.fatal("Could not instantiate communication!"+xCom.getMessage());
      throw (new xJais("Could not connect to server"));
    }
    m_worldModel=new WorldModel(f_teamName,f_isGoalie,f_playerNumber);
    m_ki=new KI(m_worldModel,m_communication,f_playerNumber);
    m_eventList=new LinkedList();
    m_singleMode=f_singleMode;
    startThread();
    try
    {
      m_communication.connectToServer();
    }
    catch(xCommunication xCom)
    {
      logger.fatal("Exception caught while connecting to server "+xCom.getMessage());
      quit();
      throw (new xJais("Could not connect to server"));
    }
    if (m_singleMode) //check if we were in singlemode, if so, startup all threads alone
    {
      startup();
      readFromCommandline();
    }
  }

  /**
   * This Method starts up our KI
   *
   * @throws xJais when something went wrong while starting up
   */
  public void startup() throws xJais
  {
    m_ki.startKI();
  }
  /**
   * @return boolean true if our thread is currently running
   * @roseuid 3CE8B146016B
   */
  public boolean isRunning()
  {
    return m_running;
  }

  /**
   * This method is called by the communication if it thinks, there is no more server present
   * @param f_timesWaited how many times it retried to receive stuff
   */
  public void informNoMoreServerPresent(int f_timesWaited)
  {
    logger.fatal("Jais #"+m_playerNumber +" says bye...lost conn to server");
    if (!m_singleMode)
      m_jaisTeam.informClientDisconnected();
    else
    {
     System.out.println("... hit RETURN to quit");
    }
    this.quit();
  }

  /**
   * Main method of JAIS
   * @param args
   * @roseuid 3CE6898C03CA
   */
  public static void main(String[] args)
  {
    FileAppender l_fa = new FileAppender();
    l_fa.setFile("jais.txt");
    l_fa.setLayout(new PatternLayout(PatternLayout.TTCC_CONVERSION_PATTERN));
    l_fa.setAppend(false);
    l_fa.activateOptions();
    logger.addAppender(l_fa);
    BasicConfigurator.configure();

    //initialize the hashtable for getting getter commandline params
    Hashtable l_filter=new Hashtable(10);
    l_filter.put("h","help");
    l_filter.put("l","loglevel");
    Hashtable l_args=CommandLineParser.Parse(args,l_filter);
    //if commandline was fine
    Jais player;
    if (l_args!=null)
    {
      logger.setLevel(Level.toLevel((String)l_args.get("loglevel")));
      System.out.println("Current logging level:" + logger.getLevel()); //check our level
      try
      {
        player=new Jais(null, (String)l_args.get("Teamname"),
                        (String)l_args.get("Servername"),
                        Integer.parseInt((String)l_args.get("Serverport")),
                        ((Boolean)l_args.get("g")).booleanValue(),
                        0,
                        true);
        } catch(xJais ex)
        {
          logger.fatal("Could not start player!"+ex.getMessage());
          return;
        }
    }
    else
    {
      Usage();
      return;
    }
    player=null;
    System.out.println("Jais says Goodbye and wishes a nice day!");
    System.exit(0);
  }

  /**
   * @param f_hearEvent
   * @roseuid 3CE8B0E601DB
   */
  public void onHear(HearEvent f_hearEvent)
  {
    synchronized(m_eventList)
    {
      m_eventList.addLast(f_hearEvent);
    }
    wakeup();
  }

  /**
   * @param f_visualEvent
   * @roseuid 3CE8B0E00179
   */
  public void onSee(VisualEvent f_visualEvent)
  {
    synchronized(m_eventList)
    {
      m_eventList.addLast(f_visualEvent);
    }
    wakeup();
  }

  /**
   *
   * @param f_initialEvent
   */
  public void onInitialEvent(InitialEvent f_initialEvent)
  {
    //get our correct PlayerNumber first...
    m_playerNumber=((InitInfo)f_initialEvent.getObjects().get(0)).m_uniformNumber;
    synchronized(m_eventList)
    {
      m_eventList.addLast(f_initialEvent);
    }
    wakeup();
  }

  /**
   * @param f_senseBodyEvent
   * @roseuid 3CE8B0D501AF
   */
  public synchronized void onSenseBody(SenseBodyEvent f_senseBodyEvent)
  {
    synchronized(m_eventList)
    {
      m_eventList.addLast(f_senseBodyEvent);
    }
    wakeup();
  }

  /**
   * @param f_kbMessage
   * @roseuid 3CE8B11602A3
   */
  public void parseKbInfo(String f_kbMessage)
  {
    StringTokenizer l_tokenizer = new StringTokenizer(f_kbMessage);
    String l_tempString = new String("");
    if (l_tokenizer.hasMoreTokens())
    {
      l_tempString=l_tokenizer.nextToken(); //get the command token
      if (l_tempString.equalsIgnoreCase("visualization") || l_tempString.equalsIgnoreCase("v"))
      {
        if (l_tokenizer.hasMoreTokens())
        {
          l_tempString=l_tokenizer.nextToken(); // get the command options
          if (l_tempString.equalsIgnoreCase("on"))
          {
            if (m_visualizationEnabled)
              logger.warn("visualization allready enabled for this player");
            else
              startVisualization();
            return;
          }
          else if (l_tempString.equalsIgnoreCase("off"))
          {
            if (m_visualizationEnabled)
              stopVisualization();
            else
              logger.warn("visualization allready disabled for this player");
            return;
          }
        }
      } //end of visualization request
      else if (l_tempString.equalsIgnoreCase("exit") || l_tempString.equalsIgnoreCase("e"))
      {
        quit();
        return;
      }
/* removed that, as it it only for hard debugging or playing ;)
        else if (l_tempString.equalsIgnoreCase("kick"))
        {
            try
            {
            m_communication.kick(Double.parseDouble(l_tokenizer.nextToken()),Double.parseDouble(l_tokenizer.nextToken()));
            }
            catch (xCommunication ex)
            {
            logger.fatal("Arg, could not send command to server!" + ex.getMessage());
            }


            return;
        }
        else if (l_tempString.equalsIgnoreCase("dash"))
        {
            try
            {
            m_communication.dash(Double.parseDouble(l_tokenizer.nextToken()));
            }
            catch (xCommunication ex)
            {
            logger.fatal("Arg, could not send command to server!" + ex.getMessage());
            }


            return;
        }
        else if (l_tempString.equalsIgnoreCase("turn"))
        {
            try
            {
            m_communication.turn(Double.parseDouble(l_tokenizer.nextToken()));
            }
            catch (xCommunication ex)
            {
            logger.fatal("Arg, could not send command to server!" + ex.getMessage());
            }
            return;
        }
        else if (l_tempString.equalsIgnoreCase("move"))
        {
            try
            {
            m_communication.move(Double.parseDouble(l_tokenizer.nextToken()),Double.parseDouble(l_tokenizer.nextToken()));
            }
            catch (xCommunication ex)
            {
            logger.fatal("Arg, could not send command to server!" + ex.getMessage());
            }

            return;
        }
*/
      else if (l_tempString.equalsIgnoreCase("help"))
      {
        System.out.println( "\nCommands are:"
                            +"\nvisualization [ON | OFF]"
                            +"\n\"Exit\" for exiting the game");
        return;
      }
    }
    logger.warn("Jais "+m_playerNumber+" didn't understand your command!");
    return;
  }

  /**
   * Method for disabling all the stuff in a correct order
   */
  public void quit()
  {
    if (m_running) stopThread();
    m_commandLineReading=false;
    if (m_visualizationEnabled)
      stopVisualization();
    try
    {
      m_communication.disconnectFromServer();
    }
    catch(xCommunication xCom)
    {
      logger.fatal("Caught exception from Communication while Jais is exiting " + xCom.getMessage());
    }

    if (m_ki!=null)
      m_ki.endTread();
    m_ki=null;
    m_worldModel = null;
  }

  /**
   * method for starting the visualization
   * @roseuid 3CE8B1020376
   */
  private void startVisualization()
  {
    m_visualization=new Visualization(m_worldModel,m_teamName, m_playerNumber);
    m_visualizationEnabled=true;
    m_visualization.createWindow();
    m_visualization.startDisplay();
  }

  /**
   * @roseuid 3CE8BFCE032D
   */
  private void stopVisualization()
  {
    m_visualization.stopDisplay();
    m_visualization.destroyWindow();
    m_visualization=null;
    m_visualizationEnabled=false;
  }

  /**
   *
   */
  private synchronized void startThread()
  {
    m_thisThread=this;
    m_running=true;
    m_thisThread.start();
  }

  /**
   * @roseuid 3CE8B0FA022A
   */
  private synchronized void stopThread()
  {
    m_thisThread = null;
    m_running=false;
    notify();
  }

  /**
   * this methods says our thread that there's something to do
   */
  private synchronized void wakeup()
  {
    m_threadSuspended=false; // ok, he can run again if he wants to
    if (!m_threadSuspended) //check if we are suspended
      notify();
  }

  /**
   * @roseuid 3CE6898C03C1
   */
  public void run()
  {
    Thread l_thisThread = currentThread();
    boolean l_listIsEmpty=true;
    Event l_tempEvent;
    while(l_thisThread==m_thisThread)
    {
      synchronized (m_eventList)
      {
        l_listIsEmpty=m_eventList.isEmpty();
      }//synchronization of m_eventlist
      if (!l_listIsEmpty) //we have something to do
      {
        try
        {
          synchronized (m_eventList)
          {
            l_tempEvent=(Event)m_eventList.removeFirst();
          }
          if (l_tempEvent instanceof SenseBodyEvent)
          {
            m_worldModel.updateSenseBodyInfo((SenseBodyEvent) l_tempEvent);
            m_ki.onSenseBody();
          }
          else if (l_tempEvent instanceof VisualEvent)
          {
            m_worldModel.updateVisualInfo((VisualEvent) l_tempEvent);
            m_ki.onNewData();
            if(m_visualizationEnabled)
              m_visualization.updateDisplay();
          }
          else if (l_tempEvent instanceof HearEvent)
          {
            m_worldModel.updateAurealInfo((HearEvent) l_tempEvent);
          }
          else if (l_tempEvent instanceof InitialEvent)
          {
            m_worldModel.updateInitInfo((InitialEvent) l_tempEvent);
          }
          else
            logger.fatal("Cannot recognize Object in Eventlist!"+l_tempEvent.getClass().getName());
        }
        catch (Exception ex)
        {
          /*        if(logger.getLevel().isGreaterOrEqual(Level.toPriority(Level.WARN_INT)))
          {
            logger.fatal("Got exception while dispatching events"+ex.getMessage());
            ex.printStackTrace();
          }
          */
        }
      }
      else
      {
        m_threadSuspended=true; //we want to suspend ourself after one run
        try
        {
          synchronized(this)
          {
            while (m_threadSuspended && l_thisThread==m_thisThread) //check again if we should sleep,
            {
              wait(); // this command lets our thread sleep until we wake him up
            }
          }
        }
        catch(java.lang.Exception ex)
        {
          logger.fatal("\t" + getName() + ": ARGG! Got an exception while sleeping ... "+ ex.getMessage());
        }
      }
    }//of while()
  }//of run()

  private void readFromCommandline()
  {
    m_inputReader = new BufferedReader(new InputStreamReader(System.in));
    String l_message= new String();
    while(m_commandLineReading)
    {
      System.out.print("\nType help for help \n>");
      try
      {
        l_message=m_inputReader.readLine();

      }
      catch (Exception ex)
      {
        logger.fatal("Exception caught in JaisTeam::readFromCommandline:" + ex.getMessage());
      }
      parseKbInfo(l_message);
    }// while()
  }//readFromCommandline
  /**
   * This method prints only the Usage of Jais to the Commandline
   * @roseuid 3CEBD85202CB
   */
  public static void Usage()
  {
    System.out.println("\nUSAGE: java Jais [TEAMNAME] [SERVERNAME] [SERVERPORT] -[OPTIONS]\n");
    System.out.println("TEAMNAME = your chosen teamname (default: Jais)");
    System.out.println("SERVERNAME = Server to connect to (default: localhost)");
    System.out.println("SERVERPORT = Port on which to connect (default: 6000)");
    System.out.println("OPTIONS = \n\t-g if I should connect as a goalie (default: no) "
                       +"\n\t-h for help "
                       +"\n\t-l NUMBER loglevel n=0..6;  0 for ALL; 6 for OFF");
    System.out.println("");
  }
}