
package jais.player;

import jais.objects.*;
import java.util.*;
import jais.world.WorldModel;
import jais.communication.*;
import org.apache.log4j.Logger;

/**
 * <p>Title: Actionmanager</p>
 * <p>Description: This class manages und stores the current actions und syncronizes the transmission to the robocup server</p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p> </p>
 * @author Christian Traub, Stephan Diederich und Claus Kestel
 */

public class ActionManager extends Thread
{

  static final double DRIBBELKICKPOWER = 0.3;

  // Params for the RunTo-Method
  private final double TOLERANCEANGLE=15;
  // Params for the kickTo-Method
  private final  double MINDISTTOBALL=0.5;



  static final int KICK = 0;
  static final int DASH = 1;
  static final int TURN = 2;
  static final int  MOVE = 3;
  static final int CATCHBALL  = 4;
  static final int SAY = 5;
  static final int NOP = 6;

  private LinkedList m_actionList;
  private WorldModel m_myWorld;
  private Communication m_myCom;
  private String m_voiceMessage;
  private int m_neckDirection=1;
  private int m_neckTurnDirection=1;
  private boolean m_visualUpdateHappened=false;

  private static Logger logger=Logger.getLogger("jais.player");
  private Thread m_performerThread;
  private boolean m_threadSuspended;
  private StateManager m_myState;
  /**
   * Creates a new <code>ActionManager</code> instance.
   *
   * @param f_myWorld a <code>WorldModel</code> value
   * @param f_myCom a <code>Communication</code> value
   * @roseuid 3CEB6AAA0062
   */
  public ActionManager(WorldModel f_myWorld, Communication f_myCom, int f_playerNumber, StateManager f_stateManager)
  {
    m_myState=f_stateManager;
    this.setName("ActionPerformer"+f_playerNumber);
    m_myWorld = f_myWorld;
    m_myCom = f_myCom;
    m_actionList = new LinkedList();
  }

  /**
   * Dribbles to an <code>Coordinate</code> in a given cycle with a given power
   *
   * @param f_power a <code>double</code> value
   * @param f_position a <code>Coordinate</code> value
   * @param f_cycle an <code>int</code> value
   * @roseuid 3CE8EF34035D
   */
  public void dribbelTo(double f_power, Coordinate f_position, int f_cycle)
  {
    switch (f_cycle){
      case 0:
        // Ball ist nah genug zu kicken
        addLastToActionList(new Action(KICK, f_power * DRIBBELKICKPOWER  ,
            m_myWorld.getBodyDirectionTo(f_position)));
        addLastToActionList(new Action(TURN, m_myWorld.getBodyDirectionTo(f_position) , 0));
        addLastToActionList(new Action(DASH, f_power , 0));
        break;
      case 1:
        nop();
        addLastToActionList(new Action(KICK, f_power * DRIBBELKICKPOWER  ,
            m_myWorld.getBodyDirectionTo(f_position)));
        addLastToActionList(new Action(TURN, m_myWorld.getBodyDirectionTo(f_position) , 0));
        break;
      case 2:
        nop();
        nop();
        addLastToActionList(new Action(KICK, f_power * DRIBBELKICKPOWER  ,
            m_myWorld.getBodyDirectionTo(f_position)));
        break;
      default:
        logger.debug("Dribbel: Cycles wrong ?!?! " );
      break;
    }
  }



  /**
   * Kicks a ball to an <code>Coordinate</code> in a given cycle with a given power
   *
   * @param f_power a <code>double</code> value
   * @param f_position a <code>Coordinate</code> value
   * @param f_cycle an <code>int</code> value
   */
  public void kickTo(double  f_power, Coordinate f_position, int f_cycle)
  {
//    logger.fatal("Kickto Cycles: " + f_cycle + " dist: " +m_myWorld.getMyDistanceTo(m_myWorld.getBall()) + "dir:" +(int)m_myWorld.getBodyDirectionTo(m_myWorld.getBall()));
    double l_turnedDirection=m_myWorld.getBodyDirectionTo(m_myWorld.getBall());
    double l_moment= l_turnedDirection  * moment();
    Coordinate l_posOfBallInNextRound=null;
    Coordinate l_posOfBallOverNextRound=null;
    Coordinate l_posOfBallOverOverNextRound=null;
    Coordinate l_posDiff=null;
    double l_dashPwr;
    double l_kickDir;
    //         logger.debug("cycle " + f_cycle);
    switch(f_cycle)
    {
      case 0:
        l_kickDir=m_myWorld.getBodyDirectionTo(f_position);
        addLastToActionList(new Action(KICK, f_power,  l_kickDir   ));
        break;
      case 1:
        // Is Ball moving?
        if (m_myWorld.getBall().getAbsPositionOfNextRound()!=null)
        {
          // Position of the ball when finally kicking
          l_posDiff=m_myWorld.getBall().getAbsPositionOfNextRound().sub(m_myWorld.getBall().getAbsPosition());
          l_dashPwr= edp2pwr(m_myWorld.getMyDistanceTo(m_myWorld.getBall().getAbsPositionOfNextRound())-MINDISTTOBALL
                             -(m_myWorld.getSenseBodyInfo().m_amountOfSpeed ));
        }
        else
        {
          //Ball is not moving, but maybe myself
          l_dashPwr= edp2pwr( m_myWorld.getMyDistanceTo( m_myWorld.getBall().getAbsPosition()) -MINDISTTOBALL
                              -m_myWorld.getSenseBodyInfo().m_amountOfSpeed
                              );
        }

        l_kickDir=m_myWorld.getBodyDirectionTo(f_position)- l_turnedDirection ;
        addLastToActionList(new Action(DASH,l_dashPwr , 0));
        addLastToActionList(new Action(KICK, f_power,  l_kickDir   ));
        break;
      case 2:
        // Is Ball moving?
        if (m_myWorld.getBall().getAbsPositionOfNextRound()!=null) {
          // Position of the ball when finally kicking
          l_posDiff=m_myWorld.getBall().getAbsPositionOfNextRound().sub(m_myWorld.getBall().getAbsPosition());
          l_posOfBallOverOverNextRound=m_myWorld.getBall().getAbsPositionOfNextRound().
                                       add(l_posDiff.scaleWith(ServerConf.BALL_DECAY)).
                                       add(l_posDiff.scaleWith(ServerConf.BALL_DECAY*ServerConf.BALL_DECAY));

          // Direction of the ball when finally kicking
          l_turnedDirection=m_myWorld.getBodyDirectionTo(l_posOfBallOverOverNextRound);
          // Moment of the ball when finally kicking
          l_moment= l_turnedDirection  * moment();
          l_dashPwr= edp2pwr(m_myWorld.getMyself().getDistanceTo(l_posOfBallOverOverNextRound)-MINDISTTOBALL
                             -(m_myWorld.getSenseBodyInfo().m_amountOfSpeed
                             *Math.pow(ServerConf.PLAYER_DECAY,3)));
        }
        else{
          //Ball is not moving, but maybe myself
          l_turnedDirection=m_myWorld.getBodyDirectionTo( m_myWorld.getBall());
          l_moment= l_turnedDirection  * moment();
          l_dashPwr= edp2pwr(m_myWorld.getBall().getDistance()-MINDISTTOBALL
                             -(m_myWorld.getSenseBodyInfo().m_amountOfSpeed* Math.pow(ServerConf.PLAYER_DECAY,3)));
        }

        l_kickDir=m_myWorld.getBodyDirectionTo(f_position)- l_turnedDirection ;
        //l_dashPwr=Math.max(0,l_dashPwr);
        l_dashPwr=Math.min(100,l_dashPwr);
        addLastToActionList(new Action(TURN, l_moment, 0));
        addLastToActionList(new Action(DASH, l_dashPwr,0));
        addLastToActionList(new Action(KICK, f_power,  l_kickDir   ));
        break;
      default:
        logger.debug("Hier hat jemand den falschen Cycle errechnet");
    }
  }

  /**
   * Kicks a ball to an <code>ObjectInfo</code> in a given cycle with a given power
   *
   * @param f_power a <code>double</code> value
   * @param f_Object an <code>ObjectInfo</code> value
   * @param f_cycle an <code>int</code> value
   * @roseuid 3CE8EF58023C
   */
  public void kickTo(double  f_power, ObjectInfo f_Object, int f_cycle)
  {
    kickTo(f_power,f_Object.getAbsPosition(),f_cycle);
  }


  /**
   * This methode is called when I'm allowed to send a command to the robocup server. It takes the first action in the actionlist, sends it to the server and removes it from the list.
   *
   * @roseuid 3CE8EFAF00A7
   */
  public void performAction()
  {
    Action l_currentAction;
    synchronized (m_actionList)
    {
      if (!m_actionList.isEmpty())
      {
        double l_actionAngle=0;
        l_currentAction=(Action)m_actionList.getFirst();
        switch(l_currentAction.getActionType())
        {
          case KICK:
            try
            {
              m_myCom.kick(l_currentAction.getActionValue1(), l_currentAction.getActionValue2());
              //turn our head to where we kicked the ball to
              l_actionAngle=l_currentAction.getActionValue2()/moment();
            }
            catch (xCommunication xCom)
            {
              logger.debug("Could not send kick action!"+xCom.getMessage());
            }

//          m_myWorld.informKick(l_currentAction.getActionValue1(), l_actionAngle  );
            break;

          case TURN:
            try
            {
              m_myCom.turn(l_currentAction.getActionValue1());
              l_actionAngle=(l_currentAction.getActionValue1() / moment());
//            nop();
            }
            catch (xCommunication xCom)
            {
              logger.debug("Could not send turn action!"+xCom.getMessage());
            }
//          m_myWorld.informTurn(l_actionAngle );
            break;

          case DASH:
            try
            {
              m_myCom.dash(l_currentAction.getActionValue1());
//            m_myWorld.informDash(l_currentAction.getActionValue1());
              if (m_myWorld.getBall().isVisible())
              {
                /** @todo turn neck here, so we dont loose ball when dashing */
              }
            }
            catch (xCommunication xCom)
            {
              logger.debug("Could not send dash action!"+xCom.getMessage());
            }

            // 	   logger.warn(" --- DASH send to Server ---") ;
            break;

          case MOVE:
            try
            {
              m_myCom.move(l_currentAction.getActionValue1(), l_currentAction.getActionValue2());
            }
            catch (xCommunication xCom)
            {
              logger.debug("Could not send move action!"+xCom.getMessage());
            }

//          m_myWorld.informMove(l_currentAction.getActionValue1(), l_currentAction.getActionValue2());
            // 	   logger.warn(" --- MOVE send to Server ---") ;
            break;
          case CATCHBALL:
            try
            {
              m_myCom.catchBall(l_currentAction.getActionValue1());
              // 		logger.warn(" --- CATCHBALL send to Server ---")   ;
            }
            catch (xCommunication xCom)
            {
              logger.debug("Could not send catch ball action!"+xCom.getMessage());
            }

//          m_myWorld.informCatchBall(m_myWorld.getBodyDirectionTo(m_myWorld.getBall()));
            break;
          case SAY:
            try
            {
              m_myCom.say(m_voiceMessage);
              // 		logger.warn(" --- SAY send to Server ---");
            }
            catch (xCommunication xCom)
            {
              logger.debug("Could not send say ball action!"+xCom.getMessage());
            }

            break;
          case NOP:
            break;
          default:
            logger.debug("Hier ist was schiefgelaufen :-(     Actionmanager.java");
        }//switch
        int l_actionType=l_currentAction.getActionType();
        if (l_actionType==ActionManager.DASH  || l_actionType==ActionManager.KICK  || l_actionType==ActionManager.TURN)
        {
          try
          {
            if (m_visualUpdateHappened && m_myWorld.getBall().isVisible())     //on first step after new see event --> turn neck to ball
            {
              double l_angleToTurnNeck=m_myWorld.getBall().getDirection()-l_actionAngle;
              if (Math.abs(l_angleToTurnNeck+m_myWorld.getNeckDirection())>90) //dont break ya neck
              {
                changeView(m_myState.m_currentViewWidth+1,true);
                m_myState.m_currentViewWidth++;
              }
              m_myCom.turnNeck(l_angleToTurnNeck);
//            m_myWorld.informTurnNeck(l_angleToBall-l_actionAngle);
              m_visualUpdateHappened=false;
            }
            else if(m_myState.m_searchingBall)
            {
              m_myCom.turnNeck(-m_myWorld.getNeckDirection());
//            m_myWorld.informTurnNeck(-m_myWorld.getNeckDirection());
            }
            else
            {
              m_myCom.turnNeck(l_actionAngle);
              if (Math.abs(l_actionAngle+m_myWorld.getNeckDirection())>90) //dont break ya neck
              {
                changeView(m_myState.m_currentViewWidth+1,true);
                m_myState.m_currentViewWidth++;
              }
//            m_myWorld.informTurnNeck(l_actionAngle);
            }
          }
          catch (xCommunication ex)
          {
            logger.debug("Caugght ex from Communication"+ex.getMessage());
          }
        }
        m_actionList.remove(0);
      }
      else //action list is empty
      {
      }
    }
  }//performAction

  /**
   * Run in the actual direction with a given power
   *
   * @param f_power an <code>int</code> value
   * @roseuid 3CE8EFC100B7
   */
  public void run(int f_power)
  {
    addLastToActionList(new Action(DASH, f_power, 0));
  }

  /**
   * Run to a <code>Coordinate</code with a given power
   *
   * @param f_power a <code>double</code> value
   * @param f_position a <code>Coordinate</code> value
   * @roseuid 3CE8EFE00170
   */
  public void runTo(double f_power,Coordinate f_position, int f_cycles)
  {
    double l_power;
    double l_deltaAngle=m_myWorld.getBodyDirectionTo(f_position);
    double l_moment= l_deltaAngle * moment();
    double l_percentageOfPower=f_power/100;
    boolean l_turnHappened=false;
//     Rckwarts rennen
//     	if (runAback(f_position)) {
//     	    f_power = -f_power;
//     	    if (l_deltaAngle > 0)
//     		l_deltaAngle -= 180;
//     	    else
//     		l_deltaAngle += 180;
//     	}

    if(Math.abs(l_deltaAngle)>TOLERANCEANGLE)
    {
      addLastToActionList(new Action(TURN,l_moment,0));
      l_turnHappened=true;
    }
    switch(f_cycles)
    {
      case SenseBodyInfo.NARROW:
        if (!l_turnHappened)
        {
          l_power=l_percentageOfPower*edp2pwr((m_myWorld.getMyDistanceTo(f_position) - ServerConf.BALL_SIZE/2-ServerConf.PLAYER_SIZE/2-m_myWorld.getSenseBodyInfo().m_amountOfSpeed));
//          logger.fatal("narrow dash sent"+Math.min(100,l_power));
          addLastToActionList(new Action(DASH,Math.min(100,l_power),0));
        }
        break;
      case SenseBodyInfo.NORMAL:
        if(l_turnHappened)
        {
          l_power=l_percentageOfPower*edp2pwr((m_myWorld.getMyDistanceTo(f_position) - ServerConf.BALL_SIZE/2-ServerConf.PLAYER_SIZE/2-m_myWorld.getSenseBodyInfo().m_amountOfSpeed));
//          logger.fatal("normal dash sent(with turn)"+Math.min(l_power,100));
          addLastToActionList(new Action(DASH,Math.min(100,l_power),0));
        }
        else
        {
          l_power=l_percentageOfPower*edp2pwr((m_myWorld.getMyDistanceTo(f_position)-ServerConf.BALL_SIZE/2-ServerConf.PLAYER_SIZE/2-m_myWorld.getSenseBodyInfo().m_amountOfSpeed));
//          logger.fatal("normal dash sent(wo turn)"+Math.min(100,l_power));
          addLastToActionList(new Action(DASH,Math.min(100,l_power/2),0));
          addLastToActionList(new Action(DASH,Math.min(100,l_power/2),0));
        }
        break;
      case SenseBodyInfo.WIDE:
        if(!l_turnHappened)
        {
          l_power=l_percentageOfPower*edp2pwr((m_myWorld.getMyDistanceTo(f_position)-ServerConf.BALL_SIZE/2-ServerConf.PLAYER_SIZE/2-m_myWorld.getSenseBodyInfo().m_amountOfSpeed));
          addLastToActionList(new Action(DASH,Math.min(100,l_power),0));
          addLastToActionList(new Action(DASH,Math.min(100,l_power),0));
          addLastToActionList(new Action(DASH,Math.min(100,l_power),0));
        }
        else
        {
          l_power=l_percentageOfPower*edp2pwr((m_myWorld.getMyDistanceTo(f_position)-ServerConf.BALL_SIZE/2-ServerConf.PLAYER_SIZE/2-m_myWorld.getSenseBodyInfo().m_amountOfSpeed));
          addLastToActionList(new Action(DASH,Math.min(100,l_power),0));
          addLastToActionList(new Action(DASH,Math.min(100,l_power),0));
        }
        break;
      default:
        return;
    }
  }

  /**
   * Wrapper for runTo with an ObjectInfo
   *
   * @param f_power a <code>double</code> value
   * @param f_Object an <code>ObjectInfo</code> value
   */
  public void runTo(double f_power,ObjectInfo f_Object, int f_cycles)
  {
    runTo(f_power,f_Object.getAbsPosition(), f_cycles);
  }


  /**
   * Turns the bodydirection of the player with a given angle
   *
   * @param f_angle a <code>double</code> value
   * @roseuid 3CE8EFF9011B
   */
  public void turn(double f_angle)
  {
    addLastToActionList(new Action(TURN, f_angle*moment(),0));
  }

  /**
   * For the goalie to catch the ball in given angle
   *
   * @param f_cycle an <code>int</code> value
   */
  public void catchTheBall(int f_cycle)
  {
    double l_deltaAngle;
    switch(f_cycle){
      case 0:
        l_deltaAngle= m_myWorld.getBodyDirectionTo(m_myWorld.getBall());
        addLastToActionList(new Action(CATCHBALL, l_deltaAngle, 0));
        break;
      case 1:
        l_deltaAngle= m_myWorld.getBodyDirectionTo( (m_myWorld.getBall().getAbsPositionOfNextRound()!=null)?m_myWorld.getBall().getAbsPositionOfNextRound():m_myWorld.getBall().getAbsPosition() );
        nop();
        addLastToActionList(new Action(CATCHBALL, l_deltaAngle, 0));
        break;
    }
  }



  /**
   * Moves the player to the given <code>Coordinate</code>
   *
   * @param f_coord a <code>Coordinate</code> value
   */
  public void moveTo(Coordinate f_coord)
  {
    addLastToActionList(new Action(MOVE, f_coord.getX(), f_coord.getY()));
  }

  /**
   * No operation
   *
   */
  public void nop()
  {
    addLastToActionList(new Action(NOP, 0,0));
  }

  private boolean runAback(Coordinate f_coord)
  {
    boolean l_back = false;
    double l_deltaAngle = 0;
    l_deltaAngle = m_myWorld.getBodyDirectionTo(f_coord);
    if (Math.abs(m_myWorld.getBodyDirectionTo(m_myWorld.getBall()) - l_deltaAngle) > 120)
      l_back = true;
    return l_back;
  }


  /**
   * Removes all elements in the Actionlist
   *
   */
  public void flush()
  {
    synchronized (m_actionList)
    {
      m_actionList.clear();
    }
  }

  /**
   * Returns <code>false</code> if the actionlist is empty
   *
   * @return a <code>boolean</code> value
   */
  public boolean listEmpty()
  {
    synchronized (m_actionList)
    {
      return (m_actionList.isEmpty());
    }
  }

  private void addLastToActionList(Action f_action)
  {
    synchronized (m_actionList)
    {
      m_actionList.add(m_actionList.size(),f_action);
    }
  }

  /**
   *
   * @param f_a
   * @return
   */
  private double getDistanceFromAccel3(double f_a){
    double l_v0=m_myWorld.getSenseBodyInfo().m_amountOfSpeed;
    double l_decay=ServerConf.PLAYER_DECAY;
    double l_v1=(l_v0+f_a)*l_decay;
    double l_v2=(l_v1+f_a)*l_decay;
    double l_distance= (l_v0+f_a)+ (l_v1+f_a)+ (l_v2+f_a);
    return l_distance;
  }

  /**
   *
   * @param f_a
   * @return
   */
  private double getDistanceFromAccel2(double f_a){
    double l_v0=m_myWorld.getSenseBodyInfo().m_amountOfSpeed;
    double l_decay=ServerConf.PLAYER_DECAY;
    double l_v1=(l_v0)*l_decay; // keine beschleunigung, da turn
    double l_v2=(l_v1+f_a)*l_decay;
    double l_distance= (l_v0)+ (l_v1+f_a)+ (l_v2+f_a);
    return l_distance;
  }


  /**
   *
   * @param f_d
   * @return
   */
  private double getAccelFromDistance3(double f_d){
    double l_v0=m_myWorld.getSenseBodyInfo().m_amountOfSpeed;
    double l_decay=ServerConf.PLAYER_DECAY;
    double l_a = (f_d - l_v0*(1+l_decay+ l_decay*l_decay))/(3+2*l_decay+ l_decay*l_decay);
    return l_a;
  }

  /**
   *
   * @param f_d
   * @return
   */
  private double getAccelFromDistance2(double f_d){
    double l_v0=m_myWorld.getSenseBodyInfo().m_amountOfSpeed;
    double l_decay=ServerConf.PLAYER_DECAY;
    double l_a = (f_d - l_v0*(1+l_decay+ l_decay*l_decay))/(2+l_decay); // die erste beschl. fehlt, da turn
    return l_a;
  }

  /**
   *
   * @param f_edp
   * @return
   */
  private double edp2pwr(double f_edp){
    return f_edp/(m_myWorld.getSenseBodyInfo().m_effort*ServerConf.DASH_POWER_RATE);
  }

  /**
   *
   * @param f_pwr
   * @return
   */
  private double pwr2edp(double f_pwr){
    return m_myWorld.getSenseBodyInfo().m_effort*ServerConf.DASH_POWER_RATE*f_pwr;
  }

  /**
   *
   * @return
   */
  private double moment(){
    return  (1+(ServerConf.INERTIA_MOMENT*m_myWorld.getSenseBodyInfo().m_amountOfSpeed));
  }



  /**
   * Returns the count of elements in the actionlist
   *
   * @return an <code>int</code> value
   */
  public int status(){
    return m_actionList.size();
  }

  /**
   * Remembers the visual update for turning the neck
   *
   */
  public void InformVisualUpdateHappened()
      {m_visualUpdateHappened=true;}

  public void changeView(int f_viewMode,boolean f_highViewQuality)
  {
    String l_viewModeString;
    if (f_viewMode==SenseBodyInfo.NARROW)
      l_viewModeString="narrow";
    else if(f_viewMode==SenseBodyInfo.WIDE)
      l_viewModeString="wide";
    else
      l_viewModeString="normal";
    try
    {
      m_myCom.changeView(l_viewModeString,f_highViewQuality);
    }
    catch (xCommunication ex)
    {
      logger.warn("changeViewQuality could not be send to server");
    }
  }

  public void run()
  {
    Thread l_thisThread = currentThread();
    while(l_thisThread==m_performerThread)
    {
      try
      {
        synchronized(this)
        {
          wait(70);
        }
        performAction();
//        logger.fatal("action performed"+System.currentTimeMillis());
      }
      catch (Exception ex)
      {
//        ex.printStackTrace();
//        logger.fatal("running on....");
      }
      m_threadSuspended=true; //we want to suspend ourself after one run
      try
      {
        synchronized(this)
        {
          while (m_threadSuspended && l_thisThread==m_performerThread) //check again if we should sleep,
          {
            wait(); // this command lets our thread sleep until we wake him up
          }
        }
      }
      catch(java.lang.Exception ex)
      {
        //logger.info("\t" + getName() + ": ARGG! Got an exception while sleeping ... "+ ex.getMessage());
      }
    }//of while()
  }

  /**
   * With this method the KI is started at the beginning
   *
   */
  public void startActionPerformer()
  {
    m_performerThread=this;
    m_performerThread.start();
  }


  /**
   * this method stops our Thread
   */
  public synchronized void stopActionPerformer()
  {
    m_performerThread = null; //get a clean outcome
    notify();
  }


  /**
   * this method says our thread that there's something to do
   */
  public synchronized void activatePerformCounter()
  {
    m_threadSuspended=false; // ok, he can run again if he wants to
    notify();
  }


}